#ifndef GLOBAL_H
#define GLOBAL_H

#include <ncurses.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

// rozmiary planszy
#define MAX_LEVEL_WIDTH 200
#define MAX_LEVEL_HEIGHT 100

// ograniczenia
#define ARRAY_SIZE(a) (sizeof(a) / sizeof(a[0]))
#define MAX_PLAYERS 4
#define MAX_INT 2147483647
#define MAX_PLAYER_MOVES 20
#define MAX_OPP_MOVES 20
#define MAX_PIGEON_MOVES 20

// czy dzialamy pod Ubuntu/Debianem?
// jesli tak trzeba dac usleep(1) po getch
#define UBUNTU 

// kierunki ruchow
#define MV_UP 0
#define MV_DOWN 1
#define MV_LEFT 2
#define MV_RIGHT 3
#define MV_STOP 4

// pakiety
#define PACKET_PLAYER_MOVES 0
#define PACKET_OPPONENT_MOVES 1
#define PACKET_PIGEON_MOVES 2
#define PACKET_POINT 3
#define PACKET_SYNCHRO 4
#define PACKET_MAP 5
#define PACKET_PLAYER_POSITION 6
#define PACKET_PIGEON_POSITION 7
#define PACKET_START_GAME 8
#define PACKET_STOP_GAME 9
#define PACKET_PLAYER_CONNECTED 10
#define PACKET_PLAYER_DISCONNECTED 11
#define PACKET_MY_NUMBER 12
#define PACKET_START_ISSUED 13
#define PACKET_STOP_ISSUED 14
#define PACKET_PIGEON_ISSUED 15
#define PACKET_PLAYER_REFUSED 16

// timestamp
#define TS_LAG -1
#define TS_INTERPOLATE -2
#define TS_WAIT -3

// rozne
#define P 1
#define G 2
#define CTRLD 	4
#define HEADING sizeof(int) + sizeof(char) + sizeof(char) + sizeof(char)

// struktury

// struktura zawierajaca timestamp i odpowiadajace mu polozenie
struct ruch
{
	int ts;	// timestamp
	short int w, k;	// polozenie
};

// struktura zawierajaca polozenie na planszy
struct pkt
{
	int w, k;	// polozenie
};

// struktura zawierajaca dane o graczach
struct players
{
	char active;	// czy pod tym numerem znajduje sie gracz?
	short int points;	// ile ma punktow?
	struct ruch buf_moves[MAX_OPP_MOVES];	// gdzie sie bedzie ruszac?
	short int buf_count;	// ile ruchow zostalo w buforze
	struct ruch buf_pred[5];	// bufor odbiorczy
	short int ile_pred;	// ile elementow w buforze odbiorczym
	struct ruch last_position;	// ostatnia znana pozycja
};

// zmienne globalne

extern struct players gracze[MAX_PLAYERS]; // dane o graczach
extern char ja;	// moj numer w tablicy "gracze"

// bufor planszy
extern char buf[MAX_LEVEL_HEIGHT][MAX_LEVEL_WIDTH];
extern int kolumny;		// ile kolumn w pliku
extern int wiersze;		// ile wierszy w pliku

extern int koniec;	// uaktywnia sie po wcisnieciu "x"
extern int dir_w;		// aktualny kierunek w pionie
extern int dir_k;		// aktualny kierunek w poziomie

extern int timestamp;	// aktualny timestamp
extern int sock;			// socket do komunikacji
extern char glag;		// golab lag indicator
extern char plag;		// player lag indicator

extern pthread_mutex_t mruch;	// mutex na bufor ruchow
extern pthread_mutex_t mtim;	// mutex na timestamp
//extern pthread_mutex_t mplay;

extern struct pkt player_pos;	// pozycja gracza
extern struct ruch buf_ruch[MAX_PLAYER_MOVES];	// ruchy gracza

extern struct ruch buf_golab[MAX_PIGEON_MOVES];	// ruchy golebia
extern struct ruch last_pigeon_position;	// ostatnia znana pozycja golebia
extern int pigeon_moves_count;	// liczba ruchow golebia w buforze

extern struct ruch tmp[5];	// tymczasowa tablica
extern int ile_ruchow;		// ile ruchow w buforze gracza
extern int akt_poz;			// aktualna pozycja w buforze gracza

extern char started;			// gra juz trwa?
extern char loading;			// plansza sie laduje?
extern char endgame;			// gra sie skonczyla?
extern char i_am_host;		// czy jestem hostem?

extern char start_issued;	// wyslano zadanie rozpoczecia gry?
extern char stop_issued;	// wyslano zadanie zakonczenia gry?
extern char pigeon_issued;	// golab zostal zlapany?

extern struct ruch zlapany; // gdzie golab zostal zlapany?

extern char servfull;	// serwer pelny?
extern char servdisc;	// serwer sie rozlaczyl?

#endif
