#include "global.h"
#include "labirynt.h"
#include "networking.h"
#include "movement.h"
#include <netinet/tcp.h>
#include <unistd.h>

// ================================================================
// funkcja main, czyli początek wszechrzeczy :>
// ================================================================
int main(int argc, char* argv[])
{
	if (argc != 2)
	{
		printf("Podaj adres IP w wywołaniu programu!\n");
		exit(1);
	}
	
	// deklaracje zmiennych od watkow
	pthread_t thk, ths, thc, th_send, th_recv;
	int idk, ids, idc, id_send, id_recv;
	
	// utworzenie socketu
	if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
	{
		perror("Blad przy tworzeniu socketu");
		return(1);
	}
	
	int tak = 1;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*) &tak, 4); 
	
	// port ustalony na 51984
	struct sockaddr_in sock_name;
	sock_name.sin_family = AF_INET;
	sock_name.sin_port = htons(51984);
	sock_name.sin_addr.s_addr = inet_addr(argv[1]);
	memset(&(sock_name.sin_zero), '\0', 8);
	
	printf("Laczenie z serwerem...\n");
	
	if ((connect(sock, (struct sockaddr *) &sock_name, sizeof(struct sockaddr))) < 0)
	{
		perror("Blad przy laczeniu");
		return(1);
	}
	
	// czyszczenie listy graczy
	int v;
	for (v=0; v<MAX_PLAYERS; v++)
	{
		gracze[v].active = 0;
		gracze[v].points = 0;
		gracze[v].buf_count = 0;
		gracze[v].ile_pred = 0;
		gracze[v].last_position.w = 0;
		gracze[v].last_position.k = 0;
	}
	
	// zaczynamy liczyc czas
	timestamp = 0;
	idc = pthread_create(&thc, NULL, count, NULL);
	
	// inicjalizacja ncurses
	initscr();
	raw();
	keypad(stdscr, TRUE);
	noecho();
	curs_set(0);
	nodelay(stdscr, TRUE);
	start_color();
	init_pair(1, COLOR_BLUE, COLOR_YELLOW);
	init_pair(2, COLOR_BLACK, COLOR_BLACK);
	init_pair(3, COLOR_WHITE, COLOR_BLACK);
	init_pair(4, COLOR_RED, COLOR_BLACK);
	init_pair(5, COLOR_WHITE, COLOR_RED);
	init_pair(6, COLOR_BLUE, COLOR_RED);
	init_pair(7, COLOR_BLACK, COLOR_GREEN);
	attrset(COLOR_PAIR(1));
	
	// inicjalizacja pthreads
	idk = pthread_create(&thk, NULL, keys, NULL); // czekanie na klawisze
	id_send = pthread_create(&th_send, NULL, socket_send, NULL); // wysylanie
	id_recv = pthread_create(&th_recv, NULL, socket_recv, NULL); // odbieranie
	ids = pthread_create(&ths, NULL, screen, NULL); // odświeżanie ekranu
	
	// czekaj na zakonczenie watkow
	pthread_join(ths, NULL);
	pthread_join(thk, NULL);
	pthread_join(thc, NULL);
	pthread_join(th_send, NULL);
	pthread_join(th_recv, NULL);
	
	// zniszcz mutexy
	pthread_mutex_destroy(&mruch);
	//pthread_mutex_destroy(&mplay);
	pthread_mutex_destroy(&mtim);
	
	// zamknij socket, ncurses i pthreads
	close(sock);
	endwin();
	pthread_exit(NULL);
	
	return 0;
}

// ================================================================
// czekanie na klawisz
// ================================================================
void* keys(void* data)
{
	// czekaj na klawisz
	char key='a';
	while (key != 'x')
	{
		
		key = getch();
		
		if (key == 'w') // gora
		{
			trymove(-1, 0);
		}
		else if (key == 's') // dol
		{
			trymove(1, 0);
		}
		else if (key == 'a') // lewo
		{
			trymove(0, -1);
		}
		else if (key == 'd') // prawo
		{
			trymove(0, 1);
		}
		else if (key == 'p')	// zarzadz start/stop gry
		{
			if (started) stop_issued = 1;
			else start_issued = 1;
		}
		
		#ifdef UBUNTU
		// bez tego pozostale watki zostaja zawieszone
		// blad wystepuje na pewno w Ubuntu 7.04, w Fedorze FC6 juz nie
		usleep(1);
		#endif
	}
	
	// nacisniete zostalo 'x'
	koniec=1;
	servdisc=0;
	servfull=0;
	shutdown(sock, SHUT_RDWR);	// zamykamy komunikacje w sockecie

	return NULL;	
}

// ================================================================
// odswiezanie planszy
// ================================================================
void* screen(void* data)
{
	int skip = 0, commit = 0;
	while (!koniec || servdisc || servfull)
	{
		skip++;
		commit++;
		
		if (skip == 2)
		{
			// co drugi ruch przesun automagicznie gracza jesli nie naciska klawiszy 
			skip = 0;
			trymove(0, 0);	// 0,0 oznacza "w ostatnio obranym kierunku"
			// tutaj ruch do bufora
		}
		
		if (commit == 1)
		{
			commit = 0;
			if (player_pos.k > kolumny) continue;
			if (player_pos.w > wiersze) continue;
			if (started)
			{
				pthread_mutex_lock(&mruch);
					
				// bufor cykliczny
				buf_ruch[akt_poz].ts = timestamp;
				buf_ruch[akt_poz].w = player_pos.w;
				buf_ruch[akt_poz].k = player_pos.k;
				akt_poz++;
				akt_poz %= MAX_PLAYER_MOVES;
				
				if (ile_ruchow < MAX_PLAYER_MOVES)
				{
					ile_ruchow++;
				}				
			pthread_mutex_unlock(&mruch);
			}
		}
		
		drawlev();			// rysuj level
		usleep(100000);	// czekaj 1/10 sekundy
	}
	return NULL;
}
	
// ================================================================
// zwieksz timestamp o 1/10 sekundy
// ================================================================
void* count(void* data)
{	
	while (!koniec)
	{
		usleep(100000);
		pthread_mutex_lock(&mtim);
		timestamp++;
		pthread_mutex_unlock(&mtim);
	}
	
	return NULL;
}

// ================================================================
// rysuj level
// ================================================================
void drawlev()
{
	
	// jesli rozlaczyl sie serwer
	if (servdisc)
	{
		attrset(COLOR_PAIR(5));
		mvprintw(10,25,"-------------------------------------");
		mvprintw(11,25,"|  Nastapilo przerwanie polaczenia  |");
		mvprintw(12,25,"|     nacisnij 'x' aby zakonczyc    |");
		mvprintw(13,25,"-------------------------------------");
		refresh();
		return;
	}
	
	// jesli nie ma dla nas miejsca
	if (servfull)
	{
		attrset(COLOR_PAIR(5));
		mvprintw(10,25,"-------------------------------------");
		mvprintw(11,25,"|           Serwer pelny!           |");
		mvprintw(12,25,"|     nacisnij 'x' aby zakonczyc    |");
		mvprintw(13,25,"-------------------------------------");
		refresh();
		return;
	}
	
	// jesli wgrywa sie plansza
	if (loading)
	{
		attrset(COLOR_PAIR(5));
		mvprintw(10,25,"-------------------------------------");
		mvprintw(11,25,"|       Wczytywanie planszy...      |");
		mvprintw(12,25,"-------------------------------------");
		refresh();
		return;
	}
	
	// ustalamy co jest widoczne z każdej strony
	int init_left = player_pos.k - 42;
	int init_top = player_pos.w - 12;
	int init_right = player_pos.k + 42;
	int init_bottom = player_pos.w + 12;
	int lev_left, lev_left_init, lev_top, lev_top_init, lev_right, lev_bottom;

	if (init_left < 0)
	{
		// w miejscu 'lev_left' pojawi się plansza
		lev_left = -1 * init_left;
		// zaczynamy rysować planszę od kolumny o indeksie '0'
		lev_left_init = 0;
	}
	else
	{
		// planszę rysujemy od początku ekranu
		lev_left = 0;
		// zaczynamy rysować planszę od kolumny 'lev_left_init'
		lev_left_init = init_left;
	}
	
	if (init_top < 0)
	{
		lev_top = -1 * init_top;
		lev_top_init = 0; 
	}
	else
	{
		lev_top = 0;
		lev_top_init = init_top;
	}
	
	if (init_right > kolumny-1)
		lev_right = lev_left + kolumny - lev_left_init;
	else
		lev_right = 84;
	
	if (init_bottom > wiersze-1)
		lev_bottom = lev_top + wiersze - lev_top_init;
	else
		lev_bottom = 24;	
	
	// gdzie jest golab???
	struct ruch pig, play[MAX_PLAYERS];
	pig = extract_move(timestamp, buf_golab, pigeon_moves_count, G);
	
	if (((pig.ts > 0) || (pig.ts == TS_INTERPOLATE)) && ((pig.w >= 0) && (pig.k >= 0) && (pig.w <= wiersze) && (pig.k <= kolumny))) 
	{
		last_pigeon_position.k = pig.k;
		last_pigeon_position.w = pig.w;
		last_pigeon_position.ts = pig.ts;
	}
	else //if ((pig.ts == TS_WAIT) || (pig.ts == TS_LAG))
	{
		pig.w = last_pigeon_position.w;
		pig.k = last_pigeon_position.k;
	}
	
	
	// gdzie sa gracze???
	int s, k, w, ak=lev_left_init, aw, already;
	for (k=0; k<MAX_PLAYERS; k++)
	{
		if ((gracze[k].active) && (k != ja))
		{
			//pthread_mutex_lock(&mplay);
			play[k] = extract_move(timestamp, gracze[k].buf_moves, gracze[k].buf_count, P);
			//pthread_mutex_unlock(&mplay);
			if (((play[k].ts > 0) || (play[k].ts == TS_INTERPOLATE)) && ((play[k].w >= 0) && (play[k].k >= 0) && (play[k].w <= wiersze) && (play[k].k <= kolumny))) 
			{
				gracze[k].last_position.k = play[k].k;
				gracze[k].last_position.w = play[k].w;
				gracze[k].last_position.ts = play[k].ts;
			}
			else// if ((play[k].ts == TS_WAIT) || (play[k].ts == TS_LAG))
			{
				play[k].w = gracze[k].last_position.w;
				play[k].k = gracze[k].last_position.k;
			}
		}
	}
	
	// rysujemy!
	
	for (k=0; k<85; k++)
	{
		aw = lev_top_init;
		for (w=0; w<24; w++)
		{
			// sprawdź czy rysować planszę, czy puste miejsce
			if ((k < lev_left) || (w < lev_top) || (k > lev_right) || (w > lev_bottom)) 
			{
				// rysujemy pustkę
				attrset(COLOR_PAIR(2));
				mvaddch(w, k, ' ');
			}
			else
			{
				// rysujemy planszę
				
				// gdzie sa gracze?
				already = 0;
				for (s=0; s<MAX_PLAYERS; s++)
				{
					if ((gracze[s].active) && (s != ja) && (play[s].w == aw) && (play[s].k == ak) && (play[s].w >= 0) && (play[s].k >= 0) && (play[s].w <= wiersze) && (play[s].k <= kolumny))
						{
						if ((play[s].ts == TS_LAG) || (play[s].ts == TS_WAIT)) attrset(COLOR_PAIR(4));
						else attrset(COLOR_PAIR(3));
						mvaddch(w, k, '&');
						already = 1;
						}
				}
				
				if (!already)
				{
					// gdzie jest golab?
					if ((pig.w == aw) && (pig.k == ak) && (pig.w >= 0) && (pig.k >= 0) && (pig.w <= wiersze) && (pig.k <= kolumny))
					{
						if ((pig.ts == TS_LAG) || (pig.ts == TS_WAIT)) attrset(COLOR_PAIR(4));
						else attrset(COLOR_PAIR(3));
						mvaddch(w, k, '*');
					}
					
					else if (buf[aw][ak] != ' ')
					{ 
						attrset(COLOR_PAIR(1));
						mvaddch(w, k, ' ');
					}
					else
					{
						attrset(COLOR_PAIR(2));
						mvaddch(w, k, ' ');
					}
				}
				aw++;
			}
		}
	if ((k >= lev_left) && (k <= lev_right))
		ak++;
	}
	
	// trzeba jeszcze zaznaczyć gdzie jesteśmy
	attrset(COLOR_PAIR(3));
	mvaddch(12, 42, '@');
	
	// teraz kolej na pasek kontrolny ...
	attrset(COLOR_PAIR(5));
	for (k=0; k<85; k++)
	{
		mvaddch(24, k, ' ');
	}
	
	mvprintw(24,2,"Punkty: ");
	k = 10;
	for (w=0; w<MAX_PLAYERS; w++)
	{
		if (gracze[w].active && (w != ja))
		{
			if (k != 10)
			{
				mvprintw(24,k,"/");
				k += 2;
			}
			mvprintw(24,k,"%d",gracze[w].points);
			k += 4;
		}
	}
	
	attrset(COLOR_PAIR(6));
	mvprintw(24,k,"Twoje punkty: %d", gracze[(int)ja].points);
	
	// ... i kontrolki podobne do tych w duzym Fiacie
	
	if (i_am_host) 
	{
		attrset(COLOR_PAIR(3));
		mvprintw(24,76,"HOST");
	}	
	
	if (plag || glag)
	{
		attrset(COLOR_PAIR(7));
		mvprintw(24,82,"LAG");
	}
	
	if (started)
	{
		attrset(COLOR_PAIR(3));
		mvprintw(24,70,"%d", timestamp);
	}
	
	// gra sie jeszcze nie zaczela
	if ((started == 0) && (endgame == 0))
	{
		attrset(COLOR_PAIR(5));
		mvprintw(10,25,"-------------------------------------");
		mvprintw(11,25,"| Oczekiwanie na rozpoczecie gry... |");
		mvprintw(12,25,"-------------------------------------");
	}
	
	// gra sie skonczyla
	else if ((started == 0) && (endgame == 1))
	{
		attrset(COLOR_PAIR(5));
		mvprintw(10,25,"-------------------------------------");
		mvprintw(11,25,"|          Gra zakonczona!          |");
		mvprintw(12,25,"-------------------------------------");	
	}
	
	refresh();
}
