#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>
#include "siec.h"
#include "defs.h"

int started = 0;

int sock; // glowne gniazdo do przyjmowania polaczen
int player[PLAYERS_MAX]; // sockety graczy

extern int generate_pigeon;
void new_pigeon_moves();

extern struct position player_pos[PLAYERS_MAX][MOVES_TO_REMEMBER];
extern int top_pos[PLAYERS_MAX];
extern int scores[PLAYERS_MAX];

extern struct position pigeon_moves[MAX_PIGEON_MOVES];
extern int timestamp;

extern char plansza[MAX_LEVEL_HEIGHT][MAX_LEVEL_WIDTH];
extern int kolumny, wiersze;

extern void random_place(short int *wiersz, short int *kolumna);
extern int last_pos;


void close_all(int)
{
	int i;

	printf("Odebrano sygnal zakonczenia serwera. Zamykanie wszystkich polaczen...\n");
	if(sock >= 0) close(sock);
	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0)
			close(player[i]);
	printf("To juz jest koniec, nie ma juz nic. Jestesmy wolni, prosimy o 5... ;)\n");
	exit(0);
}


TConnection::TConnection()
// inicjalizacja 
{
	int i;

	boss = -1;
	sock = -1;
	//pthread_mutex_lock(&mut_sock);
	for(i = 0; i < PLAYERS_MAX; i++)
		player[i] = -1;
	//pthread_mutex_unlock(&mut_sock);

	//ustawienie odpornosci na zerwanie polaczenia
	signal(SIGPIPE, SIG_IGN);
	signal(SIGINT, close_all);
}


int TConnection::create(int port)
// standardowa procedura: socket -> bind -> listen
{
	struct sockaddr_in address;
	// ustawienia dla funkcji bind().
	bzero((char*)&address, sizeof(address));
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = htonl(INADDR_ANY);
	address.sin_port = htons(port);

	// proba utworzenia gniazda
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if(sock < 0)
	{
		perror("Blad tworzenia gniazda (funkcja socket). ");
		return 1;
	}

	// przypisanie do portu
	if(bind(sock, (struct sockaddr *)&address, sizeof(address)) < 0)
	{
		perror("Blad przypisania do gniazda (funkcja bind). ");
		close(sock);
		return 2;
	}

	// oczekiwanie na jedno polaczenie
	if(listen(sock, 1) < 0)
	{
		perror("Blad funkcji listen. ");
		return 3;
	}

	// operacje na gniezdzie nie powinny byc blokujace
	fcntl(sock, F_SETFL, O_NONBLOCK);

	int tak = 1;
	setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&tak, 4);

	return 0;
}


void TConnection::host_name(int pl_sock)
{
	struct sockaddr_in inetsock;
	int length;
	char buf[20];

	length = sizeof(inetsock);
	if((getpeername(pl_sock, (struct sockaddr*)&inetsock, (socklen_t*)&length))  == 0)
	{
		strcpy(buf, inet_ntoa(inetsock.sin_addr));
		printf("Polaczenie z adresu %s\n", buf);
	}
}

void TConnection::get_something()
{
	fd_set sockets2read;
	struct timeval times;
	struct sockaddr_in pl_addr;
	int i, length = sizeof(pl_addr);
	int sel, new_sock, tak = 1;
	
	if (generate_pigeon)
	{
		new_pigeon_moves();
		generate_pigeon = 0;
	}
	
	// wyzerowanie zestwu
	FD_ZERO(&sockets2read);
	FD_SET(sock, &sockets2read);
	
	for(i = 0; i < PLAYERS_MAX; i++)
	{
		if(player[i] >= 0)
			FD_SET(player[i], &sockets2read);
	}

	times.tv_sec = 0;
	times.tv_usec = 100000l;
	sel = select(PLAYERS_MAX+2, &sockets2read, NULL, NULL, &times);
	if(sel == -1)
	{
		perror("Blad funkcji select()\n");
		return;
	}
	else if(sel == 0)
		return;

	if(FD_ISSET(sock, &sockets2read))
	{
		new_sock = accept(sock, (struct sockaddr *) &pl_addr, (socklen_t *) &length);
		if(new_sock < 0)
			perror("Blad przy wykonaniu funkcji accept()\n");
		else
		{
			for(i = 0; i < PLAYERS_MAX; i++)
			{
				if(player[i]<0)
				{
					host_name(new_sock);
					printf("Dolaczyl sie nowy gracz, otrzymuje numer %d\n", i);
					fcntl(new_sock, F_SETFL, O_NONBLOCK);
					setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, (char*)&tak, 4);
					player[i] = new_sock;
					send_init(i);
					send_new_guy(i);
					break;
				}
			}
			if(i == PLAYERS_MAX) // zbyt duzo klientow
			{
				host_name(new_sock);
				printf("Nie mozna dolaczyc nowego gracza, limit wyczerpany.\n");
				fcntl(new_sock, F_SETFL, O_NONBLOCK);
				send_too_many(new_sock);
				close(new_sock);
			}
		}
	}
	for(i = 0; i < PLAYERS_MAX; i++)
	{
		if(player[i] >= 0 && FD_ISSET(player[i], &sockets2read))
			read_sock(player[i], i);
	}
}


int TConnection::read_or_die(int pl_sock, void *buf, int length, int pl_nr)
{
	int i, got, bytes_read = 0;

	do
	{
		got = read(pl_sock, (char *)buf+bytes_read, length-bytes_read); // odczyt odpowiedniej ilosci bajtow
		if(got < 1)
		{
			if(errno != EAGAIN)
			{
				shutdown(pl_sock, SHUT_RDWR);
				player[pl_nr] = -1;
				printf("Utracono komunikacje z graczem %d, zamykanie polaczenia\n", pl_nr);
				send_player_end(pl_nr);
				boss = -1;
				if(pl_nr == boss)
				{
					for(i = 0; i < PLAYERS_MAX; i++)
					{
						if(player[i] >= 0)
						{
							boss = i;
							send_boss(i);
							break;
						}
					}
				}
				if(boss == -1) started = 0;
				return -1;
			}
		}
		else
			bytes_read += got;
		usleep(10000);
	}
	while(bytes_read < length);

	return bytes_read;
}

int TConnection::write_or_die(int pl_sock, void *buf, int length, int pl_nr)
{
	int i, num, written = 0;
	if(pl_sock < 0)
	{
		printf("Zly deskryptor gniazda w funkcji zapisujacej\n");
		return -1;
	}

	do
	{
		num = write(pl_sock, buf, length);
		if(num <= 0)
		{
			if(errno != EAGAIN)
			{
				printf("Usuwanie gracza %d (utracone polaczenie)\n", pl_nr);
				perror("");
				shutdown(player[pl_nr], SHUT_RDWR);
				close(player[pl_nr]);
				player[pl_nr] = -1;
				send_player_end(pl_nr);
				boss = -1;
				if(pl_nr == boss)
				{
					for(i = 0; i < PLAYERS_MAX; i++)
					{
						if(player[i] >= 0)
						{
							boss = i;
							send_boss(i);
							break;
						}
					}
				}
				if(boss == -1) started = 0;
				return -1;
			}
			else
				usleep(10000);
		}
		else
			written += num;
	}
	while(written < length);

	return length;
}


void TConnection::read_sock(int pl_sock, int pl_nr)
{
	char buf[1000];
	int fields, position, pack_start;
	

	if(read_or_die(pl_sock, buf, 5, pl_nr) == -1)	return; // odczyt naglowka
	
	position = 5;

	pack_start = position;

	switch(buf[position-1])	// typ pakietu
	{
		case PACKET_START_ISSUED:
			printf("Otrzymano pakiet PACKET_START_ISSUED\n");
			if(read_or_die(pl_sock, buf+position, 1, pl_nr) == -1) return;
			send_start();
			position += 1;
			send_scores();
			break;

		case PACKET_STOP_ISSUED:
			printf("Otrzymano pakiet PACKET_STOP_ISSUED\n");
			if(read_or_die(pl_sock, buf+position, 1, pl_nr) == -1) return;
			send_stop();
			position += 1;
			break;

		case PACKET_PLAYER_MOVES:
			printf("Otrzymano pakiet PACKET_PLAYER_MOVES\n");
			if(read_or_die(pl_sock, buf+position, 2, pl_nr) == -1) return; // odczyt nr gracza i ilosc ruchow
			fields = buf[position+1];
			position += 2;
			if(read_or_die(pl_sock, buf+position, 8*fields, pl_nr) == -1) return; // odczyt ruchow
			position += 8*fields;
			printf("Otrzymano %d ruchow\n", fields);
			handle_moves(buf+pack_start, position-pack_start, pl_nr);
			send_moves(buf+pack_start, position-pack_start, pl_nr);
			break;

		case PACKET_PIGEON_ISSUED:
			printf("Otrzymano pakiet PACKET_PIGEON_ISSUED\n");
			if(read_or_die(pl_sock, buf+position, 9, pl_nr) == -1) return;
			position += 9;
			handle_pigeon_issue(buf+pack_start, position-pack_start, pl_nr);
			break;

		default:
			printf("Otrzymano niepoprawny pakiet (typ %d) od gracza %d\n", buf[position-1], pl_nr);
	}
}


void TConnection::send_too_many(int refused_sock)
{
	char buf[6] = {0, 0, 0, 0, 0, 0};

	buf[4] = 1;
	buf[5] = PACKET_PLAYER_REFUSED;

	write(refused_sock, buf, 6);
	close(refused_sock);
}


void TConnection::send_new_guy(int pl_nr)
{
	int i;
	char buf[3];

	buf[0] = 1;
	buf[1] = PACKET_PLAYER_CONNECTED;
	buf[2] = (char)pl_nr;

	if(boss > -1) printf("Rozglaszanie obecnosci nowego gracza (nr %d)\n", pl_nr);
	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0 && i != pl_nr)
		{
			if(write_or_die(player[i], &timestamp, 4, pl_nr) == -1) return;
			if(write_or_die(player[i], buf, 3, pl_nr) == -1) continue;
		}
}


void TConnection::send_init(int pl_nr)
{
	char buf[5];
	int wym[2], i, usrs=0;
	short int poz_w, poz_k, pkt;
	
	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0 && i != pl_nr)
			usrs++;
	
	usrs *= 2;
	if(write_or_die(player[pl_nr], &timestamp, 4, pl_nr) == -1) return;
	buf[0] = 3+usrs;	// co najmniej trzy pola: numer gracza, plansza i polozenia poczatkowe + punkty
	if(started) buf[0] = 4+usrs; // dodatkowy pakiet startu

	buf[1] = PACKET_MY_NUMBER;
	buf[2] = (char)pl_nr;	// numer
	buf[3] = 1;		// master...
	if(boss == -1) 
	{
		buf[3] = 1;
		boss = pl_nr;
	}	
	else
		buf[3] = 0;	// ...slave
	

	buf[4] = PACKET_MAP;
	if(write_or_die(player[pl_nr], buf, 5, pl_nr) == -1) return;

	wym[0] = wiersze;
	wym[1] = kolumny;
	if(write_or_die(player[pl_nr], wym, 8, pl_nr) == -1) return;
	if(write_or_die(player[pl_nr], plansza, MAX_LEVEL_WIDTH*MAX_LEVEL_HEIGHT, pl_nr) == -1) return;

	usleep(10000);

	random_place(&poz_w, &poz_k);
	buf[0] = PACKET_PLAYER_POSITION;
	buf[1] = (char) pl_nr;
	if(write_or_die(player[pl_nr], buf, 2, pl_nr) == -1) return;
	if(write_or_die(player[pl_nr], &poz_w, 2, pl_nr) == -1) return;
	if(write_or_die(player[pl_nr], &poz_k, 2, pl_nr) == -1) return;
	
	buf[0] = PACKET_PLAYER_CONNECTED;

	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0 && i != pl_nr)
		{
			printf("Przesylanie nowemu listy obecnosci (nr %d)\n", i);
			buf[1] = (char) i;
			buf[2] = PACKET_POINT;
			buf[3] = (char) i;
			if(write_or_die(player[pl_nr], buf, 4, pl_nr) == -1) continue;
			pkt = scores[i];
			if(write_or_die(player[pl_nr], &pkt, 2, pl_nr) == -1) continue;
		}

	usleep(10000);

	if(started)
	{
		buf[0] = PACKET_START_GAME;
		if(write_or_die(player[pl_nr], buf, 1, pl_nr) == -1) return;
	}

	printf("Wyslano pakiety inicjujace do gracza %d (poz. pocz.: %d, %d)\n", pl_nr, poz_w, poz_k);
}


void TConnection::send_scores()
{
	char buf[4];
	int i, j, usrs = 0, pkt;

	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0)
			usrs++;

	printf("Rozglaszanie aktualnej punktacji\n");

	for(j = 0; j < PLAYERS_MAX; j++)
		if(player[j])
		{
			buf[0] = (char)usrs;
			if(write_or_die(player[j], &timestamp, 4, j) == -1) return;
			if(write_or_die(player[j], buf, 1, j) == -1) return;

			for(i = 0; i < PLAYERS_MAX; i++)
				if(player[i] >= 0)
				{
					buf[0] = PACKET_POINT;
					buf[1] = (char) i;
					if(write_or_die(player[j], buf, 2, j) == -1) continue;
					pkt = scores[i];
					if(write_or_die(player[j], &pkt, 2, j) == -1) continue;
				}
		}
}


void TConnection::send_boss(int pl_nr)
{
	char buf[4];
	
	if(write_or_die(player[pl_nr], &timestamp, 4, pl_nr) == -1) return;
	buf[0] = 1;
	buf[1] = PACKET_MY_NUMBER;
	buf[2] = (char)pl_nr;
	buf[3] = 1;
	if(write_or_die(player[pl_nr], buf, 4, pl_nr) == -1) return;
}


void TConnection::send_start()
{
	int i;
	char buf[2];

	if(started) return;
	buf[0] = 1;	// tylko 1 skladowa
	buf[1] = PACKET_START_GAME;

	printf("Rozpoczynanie rozgrywki\n");
	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0)
		{
			if(write_or_die(player[i], &timestamp, 4, i) == -1) continue;
			if(write_or_die(player[i], buf, 2, i) == -1) continue;
		}
	started = 1;
	bzero(scores, sizeof(short int)*PLAYERS_MAX);
}


void TConnection::send_stop()
{
	int i;
	char buf[2];

	buf[0] = 1;
	buf[1] = PACKET_STOP_GAME;

	printf("Koniec gry\n");
	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0)
		{
			if(write_or_die(player[i], &timestamp, 4, i) == -1) continue;
			if(write_or_die(player[i], buf, 2, i) == -1) continue;
		}
	started = 0;
}


void TConnection::send_player_end(int pl_nr)
{
	int i;
	char buf[3];

	buf[0] = 1;
	buf[1] = PACKET_PLAYER_DISCONNECTED;
	buf[2] = (char)pl_nr;

	if(boss > -1) printf("Rozglaszanie odlaczenia gracza %d\n", pl_nr);
	for(i = 0; i < PLAYERS_MAX; i++)
	{
		if((player[i] >= 0) && (i != pl_nr))
		{
			printf("Hej, %d, %d sie rozlaczyl!\n", i, buf[2]);
			if(write_or_die(player[i], &timestamp, 4, pl_nr) == -1) return;
			if(write_or_die(player[i], buf, 3, pl_nr) == -1) continue;
			printf(" OK!\n");
		}
	}
}


void TConnection::send_moves(char *buffer, int length, int pl_nr)
{
	int i;
	char buf[200];
	char ilep = 1;
	char pt = PACKET_OPPONENT_MOVES;

	buf[0] = 1;
	memcpy(buf, buffer, length);

	if(boss > -1) printf("Rozglaszanie ruchu gracza %d, dlug. %d\n", pl_nr, length);
	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0 && i != pl_nr)
		{
			if(write_or_die(player[i], &timestamp, 4, pl_nr) == -1) return;
			ilep = 1;
			if(write_or_die(player[i], &ilep, 1, pl_nr) == -1) return;
			if(write_or_die(player[i], &pt, 1, pl_nr) == -1) return;
			if(write_or_die(player[i], buf, length, i) == -1) continue;
		}
}


void TConnection::send_point(int pl_nr)
{
	int i;
	char buf[3];

	buf[0] = 1;
	buf[1] = PACKET_POINT;
	buf[2] = (char)pl_nr;

	if(boss > -1) printf("Rozglaszanie zdobycia punktu przez gracza %d\n", pl_nr);
	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0)
		{
			if(write_or_die(player[i], &timestamp, 4, pl_nr) == -1) return;
			if(write_or_die(player[i], buf, 3, i) == -1) continue;
			if(write_or_die(player[i], &scores[pl_nr], 2, pl_nr) == -1) return;
		}

}


void TConnection::handle_moves(char *buf, int length, int pl_nr)
{
	int i, cnt;
	int size = sizeof(player_pos[0][0]);

	buf += 2; // omijamy typ i rozmiar
	cnt = (length-2)/size;
	if(cnt < 1) return;
	if(cnt > MOVES_TO_REMEMBER)
	{
		buf += size*(cnt - MOVES_TO_REMEMBER);	// najnowsze na koncu
		cnt = MOVES_TO_REMEMBER;
	}

	for(i = 0; i < cnt; i++)
	{
		top_pos[pl_nr]++;
		top_pos[pl_nr] %= MOVES_TO_REMEMBER;
		player_pos[pl_nr][top_pos[pl_nr]].timestamp = *((int*)(buf+size*i));
		player_pos[pl_nr][top_pos[pl_nr]].y = *((short int*)(buf+size*i+4));
		player_pos[pl_nr][top_pos[pl_nr]].x = *((short int*)(buf+size*i+6));
	}
}


void TConnection::handle_pigeon_issue(char *buf, int length, int pl_nr)
{
	int min = MAX_INT;
	int i, index; 
	char ts, pl_catching;
	short int pos_x, pos_y;
	struct position *pos;

	pl_catching = buf[1];
	pos = (struct position *)(buf+2);
	
	ts = pos->timestamp;
	pos_x = pos->x;
	pos_y = pos->y;
	

	for(i = 0; i < MAX_PIGEON_MOVES; i++)
	{
		if(abs(pigeon_moves[i].timestamp - ts) < min)
		{
			min = abs(pigeon_moves[i].timestamp - ts);
			index = i;
		}
	}
	
	//if(min > MAX_TIME_DIFF) return;
	
	printf("Gracz %d zdobywa punkt!\n", pl_nr);

	scores[pl_nr]++;
	send_point(pl_nr);
	
	short poz_w, poz_k;
	random_place(&poz_w, &poz_k);
	for(i = 0; i < 2*MAX_PIGEON_MOVES; i++)
	{
		pigeon_moves[i].timestamp = timestamp;
		pigeon_moves[i].y = poz_w;
		pigeon_moves[i].x = poz_k;
	}	
}

void TConnection::new_pigeon_moves()
{
	int i, w, wb, k, kb, ts, bottom, kier, possible[4], all;

	ts = pigeon_moves[last_pos].timestamp;
	k = pigeon_moves[MAX_PIGEON_MOVES-1].x;

	for(i = 0; i < MAX_PIGEON_MOVES*2/3; i++)
	{	
		
		w = pigeon_moves[last_pos].y;
		k = pigeon_moves[last_pos].x;
		wb = pigeon_moves[(last_pos-1)%(2*MAX_PIGEON_MOVES)].y;
		kb = pigeon_moves[(last_pos-1)%(2*MAX_PIGEON_MOVES)].x;
		ts = pigeon_moves[last_pos].timestamp;
		all = 0;
			
		if((k+1 < kolumny) && (plansza[w][k+1] == ' ') && (kb != k+1))
		{
			possible[all] = 2;
			all++;
		}
		if((k-1 >= 0) && (plansza[w][k-1] == ' ') && (kb != -1))
		{
			possible[all] = 3;
			all++;
		}
		if((w+1 < wiersze) && (plansza[w+1][k] == ' ') && (wb != w+1))
		{
			possible[all] = 0;
			all++;
		}
		if((w-1 >= 0) && (plansza[w-1][k] == ' ') && (wb != w-1))
		{
			possible[all] = 1;
			all++;
		}

		if(all > 0)
		{
			kier = possible[(rand()%24) % all]; // rowne szanse
			switch(kier)
			{
				case 0: w += 1; break;
				case 1: w -= 1; break;
				case 2: k += 1; break;
				case 3: k -= 1; break;
			}
		}
		else
		{
			w = wb;
			k = kb;
		}

		last_pos = (last_pos + 1) % (2*MAX_PIGEON_MOVES);
		
		pigeon_moves[last_pos].y = w;
		pigeon_moves[last_pos].x = k;
		ts += 2;
		pigeon_moves[last_pos].timestamp = ts;
	}

	int j;
	char buf[3];
	buf[0] = 1;
	buf[1] = PACKET_PIGEON_MOVES;
	buf[2] = MAX_PIGEON_MOVES;

	if(boss > -1) printf("Wysylanie ruchow golebia\n");

	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0)
		{
			write_or_die(player[i], &timestamp, 4, i);
			write_or_die(player[i], buf, 3, i);
		}

	for(i = 0; i < PLAYERS_MAX; i++)
		if(player[i] >= 0)
		{
			for(j = 0; j < MAX_PIGEON_MOVES; j++)
			{
				bottom = (last_pos - MAX_PIGEON_MOVES + j+1); // *8
				if(bottom < 0) bottom += 2 * MAX_PIGEON_MOVES;// * 8;
				write_or_die(player[i], &(pigeon_moves[bottom]), 8, i);
			}
		}
	
}

